"""Tests for WolkConnect."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
import os
import sys
import unittest
from unittest.mock import MagicMock

sys.path.append("..")  # noqa

from wolk.wolk_connect import WolkConnect
from wolk.model.device import Device
from wolk.model.actuator_command import ActuatorCommand
from wolk.model.configuration_command import ConfigurationCommand
from wolk.model.state import State
from wolk.model.actuator_status import ActuatorStatus
from wolk.model.message import Message
from wolk.model.firmware_update_status import (
    FirmwareUpdateStatus,
    FirmwareUpdateStatusType,
)
from wolk.model.file_management_status import (
    FileManagementStatus,
    FileManagementStatusType,
)
from wolk.interface.firmware_handler import FirmwareHandler
from wolk.message_deque import MessageDeque
from wolk.wolkabout_protocol_message_deserializer import (
    WolkAboutProtocolMessageDeserializer as WAPMD,
)
from wolk.wolkabout_protocol_message_factory import (
    WolkAboutProtocolMessageFactory as WAPMF,
)
from wolk.mqtt_connectivity_service import MQTTConnectivityService


class TestWolkConnect(unittest.TestCase):
    """Tests for WolkConnect class."""

    class MockFirmwareHandler(FirmwareHandler):
        """Mock firmware installer class that whose methods will be mocked."""

        def install_firmware(self, firmware_file_path: str) -> None:
            """
            Handle the installation of the firmware file.

            :param firmware_file_path: Path where the firmware file is located
            :type firmware_file_path: str
            """
            raise NotImplementedError()

        def get_current_version(self) -> str:
            """
            Obtain device's current firmware version.

            :returns: version
            :rtype: str
            """
            raise NotImplementedError()

    def test_init_default_server(self):
        """Test creating instance with default server parameters."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)

        wolk_device = WolkConnect(device)
        self.assertIsNotNone(wolk_device.connectivity_service.ca_cert)

    def test_init_custom_server_unsecure(self):
        """Test creating instance with server on unsecure port."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)

        wolk_device = WolkConnect(device, "some_host", 1883)
        self.assertIsNone(wolk_device.connectivity_service.ca_cert)

    def test_init_custom_server_secure(self):
        """Test creating instance with server on secure port."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)

        wolk_device = WolkConnect(device, "some_host", 1883, "some_cert")
        self.assertIsNotNone(wolk_device.connectivity_service.ca_cert)

    def test_with_actuators_handler_not_callable(self):
        """Test adding faulty actuator handler that is not callable."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        self.assertRaises(ValueError, wolk_device.with_actuators, 12, 34)

    def test_with_actuators_handler_invalid_signature(self):
        """Test adding faulty actuator handler with bad signature."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        self.assertRaises(
            ValueError, wolk_device.with_actuators, lambda a, b, c: a, 34
        )

    def test_with_actuators_provider_not_callable(self):
        """Test adding faulty actuator status provider that is not callable."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        self.assertRaises(
            ValueError, wolk_device.with_actuators, lambda a, b: a, 34
        )

    def test_with_actuators_provider_invalid_signature(self):
        """Test adding actuator status provider with invalid signature."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        self.assertRaises(
            ValueError,
            wolk_device.with_actuators,
            lambda a, b: a,
            lambda a, b, c: a,
        )

    def test_with_actuators_valid(self):
        """Test adding actuator with valid handler and provider."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device).with_actuators(
            lambda a, b: a,
            lambda a: a,
        )
        self.assertIsNotNone(wolk_device.actuation_handler)

    def test_with_configuration_handler_not_callable(self):
        """Test adding configurations with handler not callable."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        self.assertRaises(
            ValueError,
            wolk_device.with_configuration,
            13,
            lambda a, b, c: a,
        )

    def test_with_configuration_handler_invalid_signature(self):
        """Test adding configurations with invalid handler signature."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        self.assertRaises(
            ValueError,
            wolk_device.with_configuration,
            lambda a, b: a,
            lambda a, b, c: a,
        )

    def test_with_configuration_provider_not_callable(self):
        """Test adding configurations with provider not callable."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        self.assertRaises(
            ValueError,
            wolk_device.with_configuration,
            lambda a: a,
            13,
        )

    def test_with_configuration_provider_invalid_signature(self):
        """Test adding configurations with invalid provider signature."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        self.assertRaises(
            ValueError,
            wolk_device.with_configuration,
            lambda a: a,
            lambda a, b, c: a,
        )

    def test_with_configuration_valid(self):
        """Test adding configurations with valid handler and provider."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device).with_configuration(
            lambda a: a, lambda: 5
        )

        self.assertIsNotNone(wolk_device.configuration_handler)

    def test_with_file_management(self):
        """Test enabling file management module."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(  # noqa
            256, 1024, file_directory
        )
        self.assertTrue(os.path.exists(file_directory))
        os.rmdir(file_directory)

    def test_with_firmware_update_no_file_management(self):
        """Test enabling firmware update module fails if no file management."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        self.assertRaises(RuntimeError, wolk_device.with_firmware_update, 12)

    def test_with_firmware_update_and_file_management(self):
        """Test enabling firmware update module with file management module."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")

        wolk_device.with_firmware_update(firmware_handler)

        firmware_handler.get_current_version.assert_called_once()

    def test_with_custom_message_queue_invalid_instance(self):
        """Test using custom message queue with passing bad message queue."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        self.assertRaises(ValueError, wolk_device.with_custom_message_queue, 1)

    def test_with_custom_message_queue_valid_instance(self):
        """Test using custom message queue with passing good message queue."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        class MockMessageQueue(MessageDeque):
            pass

        wolk_device.with_custom_message_queue(MockMessageQueue())

        self.assertIsInstance(wolk_device.message_queue, MockMessageQueue)

    def test_with_custom_protocol_bad_factory(self):
        """Test using custom protocol with bad message factory."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        self.assertRaises(ValueError, wolk_device.with_custom_protocol, 12, 34)

    def test_with_custom_protocol_bad_deserializer(self):
        """Test using custom protocol with bad message deserializer."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        class MockFactory(WAPMF):
            pass

        self.assertRaises(
            ValueError,
            wolk_device.with_custom_protocol,
            MockFactory(device_key),
            34,
        )

    def test_with_custom_protocol_valid(self):
        """Test using custom protocol with valid factory and deserializer."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        class MockFactory(WAPMF):
            pass

        class MockDeserializer(WAPMD):
            pass

        wolk_device.with_custom_protocol(
            MockFactory(device_key), MockDeserializer(device)
        )

        self.assertIsInstance(wolk_device.message_factory, MockFactory)

    def test_with_custom_connectivity_invalid_instance(self):
        """Test using custom connectivity with bad instance."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        self.assertRaises(ValueError, wolk_device.with_custom_connectivity, 1)

    def test_with_custom_connectivity_valid_instance(self):
        """Test using custom connectivity with valid instance."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        class MockCS(MQTTConnectivityService):
            pass

        mock_cs = MockCS(device, [], Message("lastwill"))
        mock_cs.set_inbound_message_listener = MagicMock()

        wolk_device.with_custom_connectivity(mock_cs)

        mock_cs.set_inbound_message_listener.assert_called_once_with(
            wolk_device._on_inbound_message
        )

    def test_connect_already_connected(self):
        """Test connecting when already connected."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.logger.info = MagicMock()

        wolk_device.connect()
        wolk_device.logger.info.assert_called_once()

    def test_connect_cs_raises_exception(self):
        """Test connecting with connectivity service failing."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        wolk_device.connectivity_service.connect = MagicMock(
            side_effect=Exception()
        )
        wolk_device.logger.exception = MagicMock()

        wolk_device.connect()
        wolk_device.logger.exception.assert_called_once()

    def test_connect_fail_to_connect(self):
        """Test connecting fails."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        wolk_device.connectivity_service.connect = MagicMock(
            return_value=False
        )
        wolk_device.logger.exception = MagicMock()

        wolk_device.connect()
        self.assertEqual(
            2, wolk_device.connectivity_service.is_connected.call_count
        )

    def test_connect_connects(self):
        """Test connecting passes."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        wolk_device.connectivity_service.is_connected = MagicMock()
        wolk_device.connectivity_service.is_connected.side_effect = [
            False,
            False,
            True,
        ]
        wolk_device.connectivity_service.connect = MagicMock(return_value=True)
        wolk_device.connect()
        self.assertEqual(
            2, wolk_device.connectivity_service.is_connected.call_count
        )

    def test_connect_publishes_actuator(self):
        """Test connecting passes and publishes actuator."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = ["A"]
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.actuator_status_provider = True
        wolk_device.publish_actuator_status = MagicMock()
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.keep_alive_service_enabled = False

        wolk_device.connectivity_service.is_connected = MagicMock()
        wolk_device.connectivity_service.is_connected.side_effect = [
            False,
            True,
        ]
        wolk_device.connectivity_service.connect = MagicMock(return_value=True)
        wolk_device.connect()
        wolk_device.publish_actuator_status.assert_called_once_with(
            actuator_references[0]
        )

    def test_connect_publishes_configuration(self):
        """Test connecting passes and publishes configuration."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.configuration_provider = True
        wolk_device.publish_configuration = MagicMock()
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.keep_alive_service_enabled = False

        wolk_device.connectivity_service.is_connected = MagicMock()
        wolk_device.connectivity_service.is_connected.side_effect = [
            False,
            True,
        ]
        wolk_device.connectivity_service.connect = MagicMock(return_value=True)
        wolk_device.connect()
        wolk_device.publish_configuration.assert_called_once()

    def test_connect_publish_file_list_fails(self):
        """Test connecting passes and fails to publishes file list."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        file_directory = "files"
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        wolk_device.file_management.get_file_list = MagicMock(return_value=[])
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.message_queue.put = MagicMock()
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.keep_alive_service_enabled = False

        wolk_device.connectivity_service.is_connected = MagicMock()
        wolk_device.connectivity_service.is_connected.side_effect = [
            False,
            True,
        ]
        wolk_device.connectivity_service.connect = MagicMock(return_value=True)
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.connect()
        wolk_device.message_queue.put.assert_called_once()
        os.rmdir(file_directory)

    def test_connect_publish_file_list(self):
        """Test connecting passes and publishes file list."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        file_directory = "files"
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        wolk_device.file_management.get_file_list = MagicMock(return_value=[])
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_queue.put = MagicMock()
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.keep_alive_service_enabled = False

        wolk_device.connectivity_service.is_connected = MagicMock()
        wolk_device.connectivity_service.is_connected.side_effect = [
            False,
            True,
        ]
        wolk_device.connectivity_service.connect = MagicMock(return_value=True)
        wolk_device.connect()
        wolk_device.message_queue.put.assert_not_called()
        os.rmdir(file_directory)

    def test_connect_publish_firmware_version_fails(self):
        """Test connecting passes and firmware version fails to send."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        file_directory = "files"
        device = Device(device_key, device_password, actuator_references)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")

        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        wolk_device.with_firmware_update(firmware_handler)
        wolk_device.file_management = None
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.message_queue.put = MagicMock()
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.keep_alive_service_enabled = False

        wolk_device.connectivity_service.is_connected = MagicMock()
        wolk_device.connectivity_service.is_connected.side_effect = [
            False,
            True,
        ]
        wolk_device.connectivity_service.connect = MagicMock(return_value=True)
        wolk_device.connect()
        wolk_device.message_queue.put.assert_called_once()
        os.rmdir(file_directory)

    def test_connect_publish_firmware_version_passes(self):
        """Test connecting passes and firmware version is sent."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        file_directory = "files"
        device = Device(device_key, device_password, actuator_references)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")

        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        wolk_device.with_firmware_update(firmware_handler)
        wolk_device.file_management = None
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_queue.put = MagicMock()
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.keep_alive_service_enabled = False

        wolk_device.connectivity_service.is_connected = MagicMock()
        wolk_device.connectivity_service.is_connected.side_effect = [
            False,
            True,
        ]
        wolk_device.connectivity_service.connect = MagicMock(return_value=True)
        wolk_device.connect()
        wolk_device.message_queue.put.assert_not_called()
        os.rmdir(file_directory)

    def test_connect_keep_alive_enabled(self):
        """Test connecting with enabled keep alive service."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)

        wolk_device = WolkConnect(device)
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_queue.put = MagicMock()
        wolk_device.logger.setLevel(logging.CRITICAL)

        wolk_device.connectivity_service.is_connected = MagicMock()
        wolk_device.connectivity_service.is_connected.side_effect = [
            False,
            True,
        ]
        wolk_device.connectivity_service.connect = MagicMock(return_value=True)
        wolk_device._send_keep_alive = MagicMock()
        wolk_device.connect()
        wolk_device._send_keep_alive.assert_called_once()
        wolk_device.keep_alive_service.cancel()

    def test_disconnect_not_connected(self):
        """Test calling disconnect when not connected."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.logger.debug = MagicMock()
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        wolk_device.disconnect()
        wolk_device.logger.debug.assert_not_called()

    def test_disconnect_when_connected(self):
        """Test calling disconnect when connected."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.logger.debug = MagicMock()
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.disconnect = MagicMock()
        wolk_device.disconnect()
        wolk_device.logger.debug.assert_called_once()

    def test_disconnect_when_connected_stops_keep_alive(self):
        """Test calling disconnect when connected stops keep alive service."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.disconnect = MagicMock()
        wolk_device.keep_alive_service = device
        wolk_device.keep_alive_service.cancel = MagicMock()
        wolk_device.disconnect()
        wolk_device.keep_alive_service.cancel.assert_called_once()

    def test_add_sensor_reading(self):
        """Test adding sensor reading to message queue."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.message_queue.put = MagicMock()
        wolk_device.add_sensor_reading("R", 1)
        wolk_device.message_queue.put.assert_called_once()

    def test_add_sensor_reading_historical(self):
        """Test adding sensor reading to message queue."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.message_queue.put = MagicMock()
        wolk_device.add_sensor_reading("R", [(1, 2), (3, 4)])
        wolk_device.message_queue.put.assert_called_once()

    def test_add_sensor_readings(self):
        """Test adding sensor readings to message queue."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.message_queue.put = MagicMock()
        wolk_device.add_sensor_readings({"R": 1})
        wolk_device.message_queue.put.assert_called_once()

    def test_add_alarm(self):
        """Test adding alarm to message queue."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.message_queue.put = MagicMock()
        wolk_device.add_alarm("R", True)
        wolk_device.message_queue.put.assert_called_once()

    def test_publish_not_connected(self):
        """Test publishing when not connected."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.logger.warning = MagicMock()
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        wolk_device.publish()
        wolk_device.logger.warning.assert_called_once()

    def test_publish_emtpy_queue(self):
        """Test publishing when queue is empty."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.message_queue.peek = MagicMock(return_value=None)
        wolk_device.publish()
        wolk_device.message_queue.peek.assert_called_once()

    def test_publish_fail_to_publish(self):
        """Test publishing and failing to publish message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.warning = MagicMock()
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.message_queue.peek = MagicMock(return_value=True)
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.publish()
        wolk_device.logger.warning.assert_called_once()

    def test_publish_success(self):
        """Test publishing successfully."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.warning = MagicMock()
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.message_queue.peek = MagicMock()
        wolk_device.message_queue.peek.side_effect = [True, None]
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_queue.get = MagicMock()
        wolk_device.publish()
        wolk_device.message_queue.get.assert_called_once()

    def test_publish_actuator_status_not_connected(self):
        """Test publishing actuator status when not connected."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.logger.warning = MagicMock()
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        wolk_device.publish_actuator_status("R")
        wolk_device.logger.warning.assert_called_once()

    def test_publish_actuator_status_no_provider(self):
        """Test publishing actuator status with no actuator status provider."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.logger.error = MagicMock()
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.publish_actuator_status("R")
        wolk_device.logger.error.assert_called_once()

    def test_publish_actuator_status_no_value(self):
        """Test publishing actuator status with no value from provider."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.logger.error = MagicMock()
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.actuator_status_provider = MagicMock(return_value=None)
        wolk_device.publish_actuator_status("R")
        wolk_device.logger.error.assert_called_once()

    def test_publish_actuator_status_no_timestamp(self):
        """Test publishing actuator status with no timestamp from provider."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        reference = "R"
        state = State.ERROR
        value = None
        actuator_status = ActuatorStatus(reference, state, value)
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.actuator_status_provider = MagicMock(
            return_value=(state, value)
        )
        wolk_device.message_factory.make_from_actuator_status = MagicMock()
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)

        wolk_device.publish_actuator_status(reference)
        wolk_device.message_factory.make_from_actuator_status.assert_called_once_with(
            actuator_status
        )

    def test_publish_actuator_status_fail_to_publish(self):
        """Test publishing actuator status fails."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        reference = "R"
        state = State.ERROR
        value = None
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.actuator_status_provider = MagicMock(
            return_value=(state, value)
        )
        wolk_device.message_factory.make_from_actuator_status = MagicMock()
        wolk_device.message_queue.put = MagicMock()
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )

        wolk_device.publish_actuator_status(reference)
        wolk_device.message_queue.put.assert_called_once()

    def test_publish_configuration_not_connected(self):
        """Test publishing configuration when not connected."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        wolk_device.logger.warning = MagicMock()
        wolk_device.publish_configuration()
        wolk_device.logger.warning.assert_called_once()

    def test_publish_configuration_no_provider(self):
        """Test publishing configuration with no configuration provider."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.logger.error = MagicMock()
        wolk_device.publish_configuration()
        wolk_device.logger.error.assert_called_once()

    def test_publish_configuration_fail_to_publish(self):
        """Test publishing configuration fails."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.configuration_provider = MagicMock()
        wolk_device.message_factory.make_from_configuration = MagicMock()
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.message_queue.put = MagicMock()
        wolk_device.publish_configuration()
        wolk_device.message_queue.put.assert_called_once()

    def test_publish_configuration_publishes(self):
        """Test publishing configuration passes."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.configuration_provider = MagicMock()
        wolk_device.message_factory.make_from_configuration = MagicMock()
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_queue.put = MagicMock()
        wolk_device.publish_configuration()
        wolk_device.message_queue.put.assert_not_called()

    def test_on_inbound_message_binary_topic(self):
        """Test on inbound message with 'binary' in topic."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("binary", "payload")
        wolk_device.logger.warning = MagicMock()

        wolk_device._on_inbound_message(message)
        wolk_device.logger.warning.assert_called_once()

    def test_on_inbound_message_unknown(self):
        """Test on inbound message for unknown message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.logger.warning = MagicMock()

        wolk_device._on_inbound_message(message)
        wolk_device.logger.warning.assert_called_once()

    def test_on_inbound_message_actuation_no_handlers(self):
        """Test on inbound actuation message but no actuation handler set."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.logger.warning = MagicMock()
        wolk_device.message_deserializer.is_actuation_command = MagicMock(
            return_value=True
        )

        wolk_device._on_inbound_message(message)
        wolk_device.logger.warning.assert_called_once()

    def test_on_inbound_message_actuation_set(self):
        """Test on inbound actuation message set message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        reference = "R"
        value = 1
        actuation_command = ActuatorCommand(reference, value)
        wolk_device.actuation_handler = MagicMock()
        wolk_device.actuator_status_provider = MagicMock()
        wolk_device.message_deserializer.parse_actuator_command = MagicMock(
            return_value=actuation_command
        )
        wolk_device.message_deserializer.is_actuation_command = MagicMock(
            return_value=True
        )
        wolk_device._on_inbound_message(message)
        wolk_device.actuation_handler.assert_called_once_with(reference, value)

    def test_on_inbound_message_configuration_no_handlers(self):
        """Test on inbound configuration message but no handler set."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.logger.warning = MagicMock()
        wolk_device.message_deserializer.is_configuration_command = MagicMock(
            return_value=True
        )
        wolk_device._on_inbound_message(message)
        wolk_device.logger.warning.assert_called_once()

    def test_on_inbound_message_configuration_set(self):
        """Test on inbound configuration message set message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        value = {"R": 1}
        configuration_command = ConfigurationCommand(value)
        wolk_device.configuration_handler = MagicMock()
        wolk_device.configuration_provider = MagicMock()
        wolk_device.message_deserializer.parse_configuration = MagicMock(
            return_value=configuration_command
        )
        wolk_device.message_deserializer.is_configuration_command = MagicMock(
            return_value=True
        )
        wolk_device._on_inbound_message(message)
        wolk_device.configuration_handler.assert_called_once_with(value)

    def test_on_inbound_message_keep_alive_response(self):
        """Test on inbound keep alive response message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)
        timestamp = 1

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.parse_keep_alive_response = MagicMock(
            return_value=timestamp
        )
        wolk_device.message_deserializer.is_keep_alive_response = MagicMock(
            return_value=True
        )
        wolk_device._on_inbound_message(message)
        self.assertEqual(timestamp, wolk_device.last_platform_timestamp)

    def test_on_inbound_message_file_management_message(self):
        """Test on inbound file management message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device._on_file_management_message = MagicMock()
        wolk_device.message_deserializer.is_file_delete_command = MagicMock(
            return_value=True
        )
        wolk_device._on_inbound_message(message)
        wolk_device._on_file_management_message.assert_called_once()

    def test_on_inbound_message_firmware_message(self):
        """Test on inbound firmware message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device._on_firmware_message = MagicMock()
        wolk_device.message_deserializer.is_firmware_install = MagicMock(
            return_value=True
        )
        wolk_device._on_inbound_message(message)
        wolk_device._on_firmware_message.assert_called_once()

    def test_on_file_management_message_no_module_fail_to_send(self):
        """Test on file management message with no module."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_factory.make_from_file_management_status = (
            MagicMock()
        )
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_management_message(message)
        wolk_device.message_queue.put.assert_called_once()

    def test_on_file_management_message_no_module_sends_error(self):
        """Test on file management message with no module, sends message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_factory.make_from_file_management_status = (
            MagicMock()
        )
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_management_message(message)
        wolk_device.message_queue.put.assert_not_called()

    def test_on_file_management_message_invalid_file_upload_init(self):
        """Test on file management message - invalid file upload initiate."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_upload_initiate = MagicMock(
            return_value=True
        )
        wolk_device.message_deserializer.parse_file_initiate = MagicMock(
            return_value=("", 0, b"")
        )
        wolk_device.file_management.handle_upload_initiation = MagicMock()
        wolk_device._on_file_management_message(message)
        wolk_device.file_management.handle_upload_initiation.assert_not_called()

    def test_on_file_management_message_file_upload_init(self):
        """Test on file management message file upload initiate."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_upload_initiate = MagicMock(
            return_value=True
        )
        wolk_device.message_deserializer.parse_file_initiate = MagicMock(
            return_value=("file", 0, b"")
        )
        wolk_device.file_management.handle_upload_initiation = MagicMock()
        wolk_device._on_file_management_message(message)
        wolk_device.file_management.handle_upload_initiation.assert_called_once_with(
            "file", 0, b""
        )

    def test_on_file_management_message_file_binary_response(self):
        """Test on file management message file binary response."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_binary_response = MagicMock(
            return_value=True
        )
        wolk_device.message_deserializer.parse_file_binary = MagicMock(
            return_value=True
        )
        wolk_device.file_management.handle_file_binary_response = MagicMock()
        wolk_device._on_file_management_message(message)
        wolk_device.file_management.handle_file_binary_response.assert_called_once_with(
            True
        )

    def test_on_file_management_message_file_upload_abort(self):
        """Test on file management message file upload abort."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_upload_abort = MagicMock(
            return_value=True
        )
        wolk_device.file_management.handle_file_upload_abort = MagicMock()
        wolk_device._on_file_management_message(message)
        wolk_device.file_management.handle_file_upload_abort.assert_called_once()

    def test_on_file_management_message_file_url_abort(self):
        """Test on file management message file url abort."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_url_abort = MagicMock(
            return_value=True
        )
        wolk_device.file_management.handle_file_upload_abort = MagicMock()
        wolk_device._on_file_management_message(message)
        wolk_device.file_management.handle_file_upload_abort.assert_called_once()

    def test_on_file_management_message_invalid_file_url_init(self):
        """Test on file management message - invalid file url initiate."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_url_initiate = MagicMock(
            return_value=True
        )
        wolk_device.message_deserializer.parse_file_url = MagicMock(
            return_value=""
        )
        wolk_device.file_management.handle_url_initiation = MagicMock()
        wolk_device._on_file_management_message(message)
        wolk_device.file_management.handle_url_initiation.assert_not_called()

    def test_on_file_management_message_file_url_init(self):
        """Test on file management message file url initiate."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_url_initiate = MagicMock(
            return_value=True
        )
        wolk_device.message_deserializer.parse_file_url = MagicMock(
            return_value="url"
        )
        wolk_device.file_management.handle_file_url_download_initiation = (
            MagicMock()
        )
        wolk_device._on_file_management_message(message)
        wolk_device.file_management.handle_file_url_download_initiation.assert_called_once_with(
            "url"
        )

    def test_on_file_management_message_file_list_request_fail_to_publish(
        self,
    ):
        """Test on file list request fails to publish and puts in queue."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_list_request = MagicMock(
            return_value=True
        )
        wolk_device.file_management.get_file_list = MagicMock(return_value=[])
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_management_message(message)

        wolk_device.message_queue.put.assert_called_once()

    def test_on_file_management_message_file_list_request_publishes(self):
        """Test on file list request fails to publish and puts in queue."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_list_request = MagicMock(
            return_value=True
        )
        wolk_device.file_management.get_file_list = MagicMock(return_value=[])
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_management_message(message)

        wolk_device.message_queue.put.assert_not_called()

    def test_on_file_management_message_file_delete_invalid_name(self):
        """Test receiving invalid file delete command."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_delete_command = MagicMock(
            return_value=True
        )
        wolk_device.message_deserializer.parse_file_delete_command = MagicMock(
            return_value=""
        )
        wolk_device.file_management.handle_file_delete = MagicMock()

        wolk_device._on_file_management_message(message)

        wolk_device.file_management.handle_file_delete.assert_not_called()

    def test_on_file_management_message_file_delete_fail_to_publish(self):
        """Test receiving file delete command and fail to publish file list."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_delete_command = MagicMock(
            return_value=True
        )
        wolk_device.message_deserializer.parse_file_delete_command = MagicMock(
            return_value="file"
        )
        wolk_device.file_management.handle_file_delete = MagicMock()
        wolk_device.file_management.get_file_list = MagicMock(return_value=[])
        wolk_device.message_factory.make_from_file_list_update = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_management_message(message)

        wolk_device.message_queue.put.assert_called_once()

    def test_on_file_management_message_file_delete_publishes(self):
        """Test receiving file delete command and send file list."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_delete_command = MagicMock(
            return_value=True
        )
        wolk_device.message_deserializer.parse_file_delete_command = MagicMock(
            return_value="file"
        )
        wolk_device.file_management.handle_file_delete = MagicMock()
        wolk_device.file_management.get_file_list = MagicMock(return_value=[])
        wolk_device.message_factory.make_from_file_list_update = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_management_message(message)

        wolk_device.message_queue.put.assert_not_called()

    def test_on_file_management_message_file_purge_fail_to_publish(self):
        """Test receiving file purge command and fail to publish file list."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_purge_command = MagicMock(
            return_value=True
        )
        wolk_device.file_management.handle_file_purge = MagicMock()
        wolk_device.file_management.get_file_list = MagicMock(return_value=[])
        wolk_device.message_factory.make_from_file_list_update = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_management_message(message)

        wolk_device.message_queue.put.assert_called_once()

    def test_on_file_management_message_file_purge_publishes(self):
        """Test receiving file purge command and send file list."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_purge_command = MagicMock(
            return_value=True
        )
        wolk_device.file_management.handle_file_purge = MagicMock()
        wolk_device.file_management.get_file_list = MagicMock(return_value=[])
        wolk_device.message_factory.make_from_file_list_update = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_management_message(message)

        wolk_device.message_queue.put.assert_not_called()

    def test_on_file_management_message_file_list_confirm(self):
        """Test receiving file list confirm message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_file_list_confirm = MagicMock(
            return_value=True
        )
        wolk_device.file_management.handle_file_list_confirm = MagicMock()

        wolk_device._on_file_management_message(message)

        wolk_device.file_management.handle_file_list_confirm.assert_called_once()

    def test_on_file_management_message_unkown(self):
        """Test receiving unknown file management message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.logger.setLevel(logging.CRITICAL)

        message = Message("some_topic", "payload")
        wolk_device.logger.warning = MagicMock()

        wolk_device._on_file_management_message(message)

        wolk_device.logger.warning.assert_called_once()

    def test_on_firmware_message_no_module_fail_to_publish(self):
        """Test receiving firmware message with no module and fail to publish."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        message = Message("some_topic", "payload")

        wolk_device.message_queue.put = MagicMock()
        wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )

        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device._on_firmware_message(message)
        wolk_device.message_queue.put.assert_called_once()

    def test_on_firmware_message_no_module_fail_publishes(self):
        """Test receiving firmware message with no module and fail to publish."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        message = Message("some_topic", "payload")
        wolk_device.logger.warning = MagicMock()

        wolk_device.message_queue.put = MagicMock()
        wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )

        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device._on_firmware_message(message)
        wolk_device.message_queue.put.assert_not_called()

    def test_on_firmware_message_firmware_install_no_path_fail_to_publish(
        self,
    ):
        """Test install command non-present file and fail to publish status."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")
        wolk_device.with_firmware_update(firmware_handler)
        wolk_device.logger.setLevel(logging.CRITICAL)
        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_firmware_install = MagicMock(
            return_value=True
        )
        wolk_device.message_deserializer.parse_firmware_install = MagicMock(
            return_value=None
        )
        wolk_device.file_management.get_file_path = MagicMock(
            return_value=None
        )
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_firmware_message(message)

        wolk_device.message_queue.put.assert_called_once()

    def test_on_firmware_message_firmware_install_no_path_publishes(
        self,
    ):
        """Test install command non-present file and publishes status."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")
        wolk_device.with_firmware_update(firmware_handler)
        wolk_device.logger.setLevel(logging.CRITICAL)
        message = Message("some_topic", "payload")
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_deserializer.is_firmware_install = MagicMock(
            return_value=True
        )
        wolk_device.message_deserializer.parse_firmware_install = MagicMock(
            return_value=None
        )
        wolk_device.file_management.get_file_path = MagicMock(
            return_value=None
        )
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_firmware_message(message)

        wolk_device.message_queue.put.assert_not_called()

    def test_on_firmware_message_firmware_install_with_path_calls_intsall(
        self,
    ):
        """Test install command present file calls handle install."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")
        wolk_device.with_firmware_update(firmware_handler)
        wolk_device.logger.setLevel(logging.CRITICAL)
        message = Message("some_topic", "payload")
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_deserializer.is_firmware_install = MagicMock(
            return_value=True
        )
        wolk_device.message_deserializer.parse_firmware_install = MagicMock(
            return_value=None
        )
        wolk_device.file_management.get_file_path = MagicMock(
            return_value="file"
        )
        wolk_device.firmware_update.handle_install = MagicMock()
        wolk_device._on_firmware_message(message)

        wolk_device.firmware_update.handle_install.assert_called_once_with(
            "file"
        )

    def test_on_firmware_message_firmware_abort(self):
        """Test abort command calls handle abort."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")
        wolk_device.with_firmware_update(firmware_handler)
        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_firmware_abort = MagicMock(
            return_value=True
        )
        wolk_device.firmware_update.handle_abort = MagicMock()
        wolk_device._on_firmware_message(message)

        wolk_device.firmware_update.handle_abort.assert_called_once_with()

    def test_on_firmware_message_firmware_version_request_pulishes(self):
        """Test receiving version request and publishes response."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")
        wolk_device.with_firmware_update(firmware_handler)
        message = Message("some_topic", "payload")
        wolk_device.message_deserializer.is_firmware_version_request = (
            MagicMock(return_value=True)
        )
        wolk_device.message_factory.make_from_firmware_version_response = (
            MagicMock(return_value="1.0")
        )
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_firmware_message(message)

        wolk_device.message_queue.put.assert_not_called()

    def test_on_firmware_message_unknown(
        self,
    ):
        """Test receiving unknown firmware message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")
        wolk_device.with_firmware_update(firmware_handler)
        message = Message("some_topic", "payload")
        wolk_device.logger.warning = MagicMock()

        wolk_device._on_firmware_message(message)

        wolk_device.logger.warning.assert_called_once()

    def test_on_package_request_fails_to_publish(self):
        """Test making package request and failing to send it."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.message_factory.make_from_package_request = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_package_request("file", 0, 64)

        wolk_device.message_queue.put.assert_called_once()

    def test_on_package_request_publishes(self):
        """Test making package request and publishing it."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        wolk_device.message_factory.make_from_package_request = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_package_request("file", 0, 64)

        wolk_device.message_queue.put.assert_not_called()

    def test_on_firmware_update_status_not_connected(self):
        """Test on firmware update status call when not connected."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")
        wolk_device.with_firmware_update(firmware_handler)

        status = FirmwareUpdateStatus(FirmwareUpdateStatusType.INSTALLING)
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        wolk_device.connectivity_service.publish = MagicMock()
        wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )
        wolk_device._on_firmware_update_status(status)

        wolk_device.connectivity_service.publish.assert_not_called()

    def test_on_firmware_update_status_fail_to_publish(self):
        """Test on firmware update status and fail to publish."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")
        wolk_device.with_firmware_update(firmware_handler)

        status = FirmwareUpdateStatus(FirmwareUpdateStatusType.INSTALLING)
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )
        wolk_device.message_queue.put = MagicMock()
        wolk_device._on_firmware_update_status(status)

        wolk_device.message_queue.put.assert_called_once()

    def test_on_firmware_update_status_publishes(self):
        """Test on firmware update status and publishes the message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")
        wolk_device.with_firmware_update(firmware_handler)

        status = FirmwareUpdateStatus(FirmwareUpdateStatusType.INSTALLING)
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )
        wolk_device.message_queue.put = MagicMock()
        wolk_device._on_firmware_update_status(status)

        wolk_device.message_queue.put.assert_not_called()

    def test_on_firmware_update_status_completed_not_connected(self):
        """Test on firmware status completed and not connected."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")
        wolk_device.with_firmware_update(firmware_handler)

        status = FirmwareUpdateStatus(FirmwareUpdateStatusType.COMPLETED)
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        wolk_device.connectivity_service.publish = MagicMock()
        wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )
        wolk_device.message_factory.make_from_firmware_version_update = (
            MagicMock(return_value=True)
        )
        wolk_device._on_firmware_update_status(status)

        wolk_device.connectivity_service.publish.assert_not_called()

    def test_on_firmware_update_status_completed_fail_to_publish(self):
        """Test on firmware status completed and fail to publish."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")
        wolk_device.with_firmware_update(firmware_handler)

        status = FirmwareUpdateStatus(FirmwareUpdateStatusType.COMPLETED)
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )
        wolk_device.message_factory.make_from_firmware_version_update = (
            MagicMock(return_value=True)
        )
        wolk_device.message_queue.put = MagicMock()
        wolk_device._on_firmware_update_status(status)

        self.assertEqual(2, wolk_device.message_queue.put.call_count)

    def test_on_firmware_update_status_completed_publishes(self):
        """Test on firmware status completed and publishes message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)
        firmware_handler = self.MockFirmwareHandler()
        firmware_handler.get_current_version = MagicMock(return_value="1.0")
        wolk_device.with_firmware_update(firmware_handler)

        status = FirmwareUpdateStatus(FirmwareUpdateStatusType.COMPLETED)
        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_factory.make_from_firmware_update_status = (
            MagicMock(return_value=True)
        )
        wolk_device.message_factory.make_from_firmware_version_update = (
            MagicMock(return_value=True)
        )
        wolk_device.message_queue.put = MagicMock()
        wolk_device._on_firmware_update_status(status)
        wolk_device.message_queue.put.assert_not_called()

    def test_on_file_upload_status_fail_to_publish(self):
        """Test on file upload status and fail to publish message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)

        wolk_device.message_factory.make_from_file_management_status = (
            MagicMock(return_value=True)
        )
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        status = FileManagementStatus(FileManagementStatusType.FILE_TRANSFER)
        file_name = "file"
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_upload_status(file_name, status)

        wolk_device.message_queue.put.assert_called_once()

    def test_on_file_upload_status_publishes(self):
        """Test on file upload status and publishes message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)

        wolk_device.message_factory.make_from_file_management_status = (
            MagicMock(return_value=True)
        )
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        status = FileManagementStatus(FileManagementStatusType.FILE_TRANSFER)
        file_name = "file"
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_upload_status(file_name, status)

        wolk_device.message_queue.put.assert_not_called()

    def test_on_file_upload_status_file_ready_fail_to_publish(self):
        """Test on file upload status and fail to publish message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)

        wolk_device.message_factory.make_from_file_management_status = (
            MagicMock(return_value=True)
        )
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.message_factory.make_from_file_list_update = MagicMock(
            return_value=True
        )
        wolk_device.file_management.get_file_list = MagicMock()
        status = FileManagementStatus(FileManagementStatusType.FILE_READY)
        file_name = "file"
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_upload_status(file_name, status)

        self.assertEqual(2, wolk_device.message_queue.put.call_count)

    def test_on_file_upload_status_file_ready_published(self):
        """Test on file upload status and publishes message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)

        wolk_device.message_factory.make_from_file_management_status = (
            MagicMock(return_value=True)
        )
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_factory.make_from_file_list_update = MagicMock(
            return_value=True
        )
        wolk_device.file_management.get_file_list = MagicMock()
        status = FileManagementStatus(FileManagementStatusType.FILE_READY)
        file_name = "file"
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_upload_status(file_name, status)

        wolk_device.message_queue.put.assert_not_called()

    def test_on_file_url_status_fail_to_publish(self):
        """Test on file url status and fail to publish update."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)

        wolk_device.message_factory.make_from_file_url_status = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        status = FileManagementStatus(FileManagementStatusType.FILE_TRANSFER)
        file_url = "file_url"
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_url_status(file_url, status)

        wolk_device.message_queue.put.assert_called_once()

    def test_on_file_url_status_publishes(self):
        """Test on file url status and publishes update."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)

        wolk_device.message_factory.make_from_file_url_status = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        status = FileManagementStatus(FileManagementStatusType.FILE_TRANSFER)
        file_url = "file_url"
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_url_status(file_url, status)

        wolk_device.message_queue.put.assert_not_called()

    def test_on_file_url_status_with_file_name_fail_to_publish(self):
        """Test on url upload status and fail to publish message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)

        wolk_device.message_factory.make_from_file_management_status = (
            MagicMock(return_value=True)
        )
        wolk_device.connectivity_service.publish = MagicMock(
            return_value=False
        )
        wolk_device.message_factory.make_from_file_url_status = MagicMock(
            return_value=True
        )
        wolk_device.file_management.get_file_list = MagicMock()
        status = FileManagementStatus(FileManagementStatusType.FILE_READY)
        file_name = "file"
        file_url = "file_url"
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_url_status(file_url, status, file_name)

        self.assertEqual(2, wolk_device.message_queue.put.call_count)

    def test_on_file_url_status_with_file_name_publishes(self):
        """Test on url upload status and publishes message."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        file_directory = "files"
        wolk_device = WolkConnect(device).with_file_management(
            256, 1024, file_directory
        )
        os.rmdir(file_directory)

        wolk_device.message_factory.make_from_file_management_status = (
            MagicMock(return_value=True)
        )
        wolk_device.connectivity_service.publish = MagicMock(return_value=True)
        wolk_device.message_factory.make_from_file_url_status = MagicMock(
            return_value=True
        )
        wolk_device.file_management.get_file_list = MagicMock()
        status = FileManagementStatus(FileManagementStatusType.FILE_READY)
        file_name = "file"
        file_url = "file_url"
        wolk_device.message_queue.put = MagicMock()

        wolk_device._on_file_url_status(file_url, status, file_name)

        wolk_device.message_queue.put.assert_not_called()

    def test_keep_alive_service_disabled(self):
        """Test keep alive service is disabled."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)

        wolk_device = WolkConnect(device).with_keep_alive_service(False)

        self.assertFalse(wolk_device.keep_alive_service_enabled)
        self.assertIsNone(wolk_device.keep_alive_service)

    def test_keep_alive_service_custom_interval(self):
        """Test on url upload status and publishes message.."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        interval = 15

        wolk_device = WolkConnect(device).with_keep_alive_service(
            True, interval
        )

        self.assertTrue(wolk_device.keep_alive_service_enabled)
        self.assertEqual(interval, wolk_device.keep_alive_interval)

    def test_send_keep_alive_not_connected(self):
        """Test seding keep alive message when not connected."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=False
        )
        wolk_device.message_factory.make_keep_alive_message = MagicMock()

        wolk_device._send_keep_alive()

        wolk_device.message_factory.make_keep_alive_message.assert_not_called()

    def test_send_keep_alive_connected(self):
        """Test seding keep alive message when connected."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)

        wolk_device.connectivity_service.is_connected = MagicMock(
            return_value=True
        )
        wolk_device.connectivity_service.publish = MagicMock()

        wolk_device._send_keep_alive()

        wolk_device.connectivity_service.publish.assert_called_once()

    def test_request_timestamp_when_none(self):
        """Test request timestamp returns none."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)

        self.assertIsNone(wolk_device.request_timestamp())

    def test_request_timestamp_when_not_none(self):
        """Test request timestamp returns value."""
        device_key = "some_key"
        device_password = "some_password"
        actuator_references = []
        device = Device(device_key, device_password, actuator_references)
        wolk_device = WolkConnect(device)
        wolk_device.logger.setLevel(logging.CRITICAL)
        wolk_device.last_platform_timestamp = 1

        self.assertIsNotNone(wolk_device.request_timestamp())



"""Process messages received from WolkAbout IoT Platform."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import List
from typing import Tuple

from wolk.model.actuator_command import ActuatorCommand
from wolk.model.configuration_command import ConfigurationCommand
from wolk.model.file_transfer_package import FileTransferPackage
from wolk.model.message import Message


class MessageDeserializer(ABC):
    """Deserialize messages received from the platform."""

    @abstractmethod
    def get_inbound_topics(self) -> List[str]:
        """
        Return list of inbound topics for device.

        :returns: List of topics to subscribe to
        :rtype: List[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def is_actuation_command(self, message: Message) -> bool:
        """
        Check if message is actuation command.

        :param message: The message received
        :type message: Message
        :returns: actuation_command
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_keep_alive_response(self, message: Message) -> bool:
        """
        Check if message is keep alive response.

        :param message: The message received
        :type message: Message
        :returns: keep_alive_response
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_configuration_command(self, message: Message) -> bool:
        """
        Check if message is configuration command.

        :param message: The message received
        :type message: Message
        :returns: configuration_command
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_firmware_install(self, message: Message) -> bool:
        """
        Check if message is firmware update install command.

        :param message: The message received
        :type message: Message
        :returns: firmware_update_install_command
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_firmware_abort(self, message: Message) -> bool:
        """
        Check if message is firmware update command.

        :param message: The message received
        :type message: Message
        :returns: firmware_update_abort_command
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_binary_response(self, message: Message) -> bool:
        """
        Check if message is file binary message.

        :param message: The message received
        :type message: Message
        :returns: file_binary_response
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_delete_command(self, message: Message) -> bool:
        """
        Check if message if file delete command.

        :param message: The message received
        :type message: Message
        :returns: file_delete_command
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_purge_command(self, message: Message) -> bool:
        """
        Check if message if file purge command.

        :param message: The message received
        :type message: Message
        :returns: file_purge_command
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_list_confirm(self, message: Message) -> bool:
        """
        Check if message is file list confirm.

        :param message: The message received
        :type message: Message
        :returns: file_list_confirm
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_list_request(self, message: Message) -> bool:
        """
        Check if message is file list request.

        :param message: The message received
        :type message: Message
        :returns: file_list_request
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_upload_initiate(self, message: Message) -> bool:
        """
        Check if message is file upload command.

        :param message: The message received
        :type message: Message
        :returns: file_upload_initiate_command
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_upload_abort(self, message: Message) -> bool:
        """
        Check if message is file upload command.

        :param message: The message received
        :type message: Message
        :returns: file_upload_abort_command
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_url_initiate(self, message: Message) -> bool:
        """
        Check if message is file URL download command.

        :param message: The message received
        :type message: Message
        :returns: file_url_download_initiate
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_url_abort(self, message: Message) -> bool:
        """
        Check if message is file URL download command.

        :param message: The message received
        :type message: Message
        :returns: file_url_download_abort
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_actuator_command(self, message: Message) -> ActuatorCommand:
        """
        Parse the message into an ActuatorCommand.

        :param message: The message received
        :type message: Message
        :returns: actuation
        :rtype: ActuatorCommand
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_keep_alive_response(self, message: Message) -> int:
        """
        Parse the message into an UTC timestamp.

        :param message: The message received
        :type message: Message
        :returns: timestamp
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_firmware_install(self, message: Message) -> str:
        """
        Return file name from message.

        :param message: The message received
        :type message: Message
        :returns: file_name
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_file_initiate(self, message: Message) -> Tuple[str, int, str]:
        """
        Return file name, file size and file hash from message.

        :param message: The message received
        :type message: Message
        :returns: (file_name, file_size, file_hash)
        :rtype: Tuple[str, int, str]
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_file_url(self, message: Message) -> str:
        """
        Parse the message into a URL string.

        :param message: The message received
        :type message: Message
        :returns file_url:
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_file_binary(self, message: Message) -> FileTransferPackage:
        """
        Parse the message into a file transfer package.

        :param message: The message received
        :type message: Message
        :returns: file_transfer_package
        :rtype: FileTransferPackage
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_configuration(self, message: Message) -> ConfigurationCommand:
        """
        Parse the message into a ConfigurationCommand.

        :param message: The message received
        :type message: Message
        :returns: configuration
        :rtype: ConfigurationCommand
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_file_delete_command(self, message: Message) -> str:
        """
        Parse the message into a file name to delete.

        :param message: The message received
        :type message: Message
        :returns: file_name
        :rtype: str
        """
        raise NotImplementedError()



"""Tests for OSFileManagement."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
import os
import sys
import unittest
from tempfile import NamedTemporaryFile
from threading import Timer
from unittest.mock import MagicMock

sys.path.append("..")  # noqa

from wolk.os_file_management import OSFileManagement
from wolk.model.file_management_error_type import FileManagementErrorType
from wolk.model.file_management_status import FileManagementStatus
from wolk.model.file_management_status_type import FileManagementStatusType
from wolk.model.file_transfer_package import FileTransferPackage


class TestOSFileManagement(unittest.TestCase):
    """Tests for OSFileManagement class."""

    def test_configure_no_existing_folder(self):
        """Test configuring file management module and create files folder."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )

        preferred_package_size = 1000
        max_file_size = 1000000
        file_directory = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "files"
        )
        file_management.configure(
            preferred_package_size, max_file_size, file_directory
        )
        self.assertTrue(os.path.exists(file_directory))
        os.rmdir(file_directory)

    def test_configure_existing_folder(self):
        """Test configuring file management module with existing folder."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        preferred_package_size = 1000
        max_file_size = 1000000
        file_directory = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "files"
        )
        os.makedirs(os.path.abspath(file_directory))
        file_management.configure(
            preferred_package_size, max_file_size, file_directory
        )
        self.assertTrue(os.path.exists(file_directory))
        os.rmdir(file_directory)

    def test_handle_upload_initiation_not_idle_state(self):
        """Test handle upload initiation when module not idle."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)

        file_name = "file"
        file_size = 1024
        file_hash = "some_hash"
        file_management.current_status = 1  # Not None

        file_management.handle_upload_initiation(
            file_name, file_size, file_hash
        )

        file_management.status_callback.assert_not_called()

    def test_handle_upload_initiation_unconfigured(self):
        """Test handle upload initiation when module not configured."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)

        file_name = "file"
        file_size = 1024
        file_hash = "some_hash"

        file_management.handle_upload_initiation(
            file_name, file_size, file_hash
        )

        expected_status = FileManagementStatus(
            FileManagementStatusType.ERROR,
            FileManagementErrorType.TRANSFER_PROTOCOL_DISABLED,
        )

        file_management.status_callback.assert_called_once_with(
            file_name, expected_status
        )

    def test_handle_upload_initiation_file_too_big(self):
        """Test handle upload initiation when file too big."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        preferred_package_size = 256
        max_file_size = 512
        file_directory = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "files"
        )
        os.makedirs(os.path.abspath(file_directory))
        file_management.configure(
            preferred_package_size, max_file_size, file_directory
        )

        file_name = "file"
        file_size = 1024
        file_hash = "some_hash"

        file_management.handle_upload_initiation(
            file_name, file_size, file_hash
        )

        expected_status = FileManagementStatus(
            FileManagementStatusType.ERROR,
            FileManagementErrorType.UNSUPPORTED_FILE_SIZE,
        )

        file_management.status_callback.assert_called_once_with(
            file_name, expected_status
        )
        os.rmdir(file_directory)

    def test_handle_upload_initiation_valid_file(self):
        """Test handle upload initiation for valid file."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        preferred_package_size = 256
        max_file_size = 1024
        file_directory = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "files"
        )
        os.makedirs(os.path.abspath(file_directory))
        file_management.configure(
            preferred_package_size, max_file_size, file_directory
        )

        file_name = "file"
        file_size = 512
        file_hash = "some_hash"

        file_management.handle_upload_initiation(
            file_name, file_size, file_hash
        )

        expected_status = FileManagementStatus(
            FileManagementStatusType.FILE_TRANSFER
        )

        file_management.status_callback.assert_called_once_with(
            file_name, expected_status
        )
        os.rmdir(file_directory)
        file_management.request_timeout.cancel()
        file_management.temp_file.close()

    def test_handle_upload_initiation_small_file(self):
        """Test handle upload initiation for small file."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        preferred_package_size = 512
        max_file_size = 1024
        file_directory = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "files"
        )
        os.makedirs(os.path.abspath(file_directory))
        file_management.configure(
            preferred_package_size, max_file_size, file_directory
        )

        file_name = "file"
        file_size = 256
        file_hash = "some_hash"

        file_management.handle_upload_initiation(
            file_name, file_size, file_hash
        )

        file_management.packet_request_callback.assert_called_once_with(
            file_name, 0, file_size + 64
        )
        os.rmdir(file_directory)
        file_management.request_timeout.cancel()
        file_management.temp_file.close()

    def test_handle_abort_with_temp_file(self):
        """Test aborting file transfer with temp file set."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        file_management.temp_file = NamedTemporaryFile(
            mode="a+b", delete=False
        )

        file_management.handle_file_upload_abort()

        self.assertIsNone(file_management.temp_file)

    def test_handle_abort(self):
        """Test aborting file transfer."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)

        file_management.handle_file_upload_abort()

        self.assertIsNone(file_management.temp_file)

    def test_file_package_binary_idle_state(self):
        """Test receiving file package when in idle state."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        file_management.logger.warning = MagicMock()

        file_transfer_package = FileTransferPackage(b"", b"", b"")

        file_management.handle_file_binary_response(file_transfer_package)

        file_management.logger.warning.assert_called_once()

    def test_file_package_binary_cancel_timeout(self):
        """Test receiving file package cancels request timeout timer."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        file_management.current_status = True
        file_management.retry_count = 0
        file_management.request_timeout = Timer(60.0, file_management._timeout)
        file_management.request_timeout.cancel = MagicMock()

        file_transfer_package = FileTransferPackage(b"", b"", b"")

        file_management.handle_file_binary_response(file_transfer_package)
        file_management.request_timeout.cancel()
        file_management.packet_request_callback.assert_called()

    # def test_file_package_binary(self):
    #     """Test receiving file package cancels request timeout timer."""
    #     mock_status_callback = MagicMock(return_value=None)
    #     mock_packet_request_callback = MagicMock(return_value=None)
    #     mock_url_status_callback = MagicMock(return_value=None)

    #     file_management = OSFileManagement(
    #         mock_status_callback,
    #         mock_packet_request_callback,
    #         mock_url_status_callback,
    #     )
    #     file_management.logger.setLevel(logging.CRITICAL)
    #     file_management.current_status = True
    #     file_management.retry_count = 0
    #     data = b"Let's try something else"
    #     current_hash = hashlib.sha256(data).digest()

    #     hashlib.sha256 = MagicMock(return_value=b"")

    #     file_transfer_package = FileTransferPackage(
    #         32 * b"\x00", data, current_hash
    #     )

    #     file_management.handle_file_binary_response(file_transfer_package)
    #     file_management.request_timeout.cancel()
    #     file_management.packet_request_callback.assert_called()

    def test_handle_file_url_download_abort(self):
        """Test method resets state."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        file_management.handle_file_url_download_abort()
        self.assertIsNone(file_management.current_status)

    def test_get_file_list_current_dir(self):
        """Test get file list for running in current directory."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        file_list = file_management.get_file_list()
        self.assertNotEqual(0, len(file_list))

    def test_get_file_list_empty_dir(self):
        """Test get file list for running in current directory."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        os.mkdir("test_dir")
        file_management.file_directory = "test_dir"
        file_management.logger.setLevel(logging.CRITICAL)
        file_list = file_management.get_file_list()
        self.assertEqual(0, len(file_list))
        os.rmdir("test_dir")

    def test_get_file_path_existing(self):
        """Test get file path for existing file."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        file_name = "test_file"
        file_handle = open(file_name, "w")
        file_handle.close()
        file_path = file_management.get_file_path(file_name)
        expected_file_path = os.path.join(
            os.path.abspath(os.getcwd()), file_name
        )
        self.assertEqual(expected_file_path, file_path)
        os.remove(file_name)

    def test_get_file_path_non_existing(self):
        """Test get file path for non_existing file."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        file_name = "test_file"
        file_path = file_management.get_file_path(file_name)
        self.assertIsNone(file_path)

    def test_handle_file_list_confirm_does_nothing(self):
        """Test file list confirm doesn't call status callbacks."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        file_management.handle_file_list_confirm()
        file_management.status_callback.assert_not_called()

    def test_handle_file_delete_non_existing(self):
        """Test deleting file that doesn't exist."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        file_name = "test_file"
        file_management.handle_file_delete(file_name)
        self.assertFalse(os.path.exists(os.path.join(os.getcwd(), file_name)))

    def test_handle_file_delete_existing(self):
        """Test deleting file that exists."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        file_name = "test_file"
        file_handle = open(file_name, "w")
        file_handle.close()
        file_management.handle_file_delete(file_name)
        self.assertFalse(os.path.exists(os.path.join(os.getcwd(), file_name)))

    def test_handle_file_purge(self):
        """Test deleting all regular files in a directory."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        file_names = ["file1", "file2", ".special-file"]
        files_directory = "test_dir"
        os.mkdir(files_directory)
        directory_path = os.path.join(os.getcwd(), files_directory)
        for file in file_names:
            file_handle = open(os.path.join(directory_path, file), "w")
            file_handle.close()

        file_management.file_directory = directory_path
        file_management.handle_file_purge()
        self.assertEqual(1, len(os.listdir(directory_path)))
        os.remove(os.path.join(directory_path, ".special-file"))
        os.rmdir(files_directory)

    def test_timeout(self):
        """Test timeout calls abort."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        file_management.handle_file_upload_abort = MagicMock()

        file_management._timeout()
        file_management.handle_file_upload_abort.assert_called_once()

    def test_handle_file_url_download_init_not_idle(self):
        """Test URL upload init when not in idle state."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        file_management.logger.warning = MagicMock()
        file_management.current_status = True

        file_management.handle_file_url_download_initiation("some_url")
        file_management.logger.warning.assert_called_once()

    def test_handle_file_url_download_init_invalid_url(self):
        """Test URL upload init for invalid url."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        file_management.logger.setLevel(logging.CRITICAL)
        file_management.logger.error = MagicMock()

        file_management.handle_file_url_download_initiation("some_url")
        file_management.logger.error.assert_called_once()

    def test_handle_file_url_download_init_valid_url(self):
        """Test URL upload init for valid url."""
        mock_status_callback = MagicMock(return_value=None)
        mock_packet_request_callback = MagicMock(return_value=None)
        mock_url_status_callback = MagicMock(return_value=None)

        file_management = OSFileManagement(
            mock_status_callback,
            mock_packet_request_callback,
            mock_url_status_callback,
        )
        os.mkdir("test_dir")
        file_management.file_directory = "test_dir"
        file_management.logger.setLevel(logging.CRITICAL)
        file_management.logger.error = MagicMock()

        licence_url = (
            "https://raw.githubusercontent.com/Wolkabout"
            + "/WolkConnect-Python/master/LICENSE"
        )

        file_management.handle_file_url_download_initiation(licence_url)
        status = FileManagementStatus(FileManagementStatusType.FILE_READY)
        file_management.url_status_callback.assert_called_with(
            licence_url, status, "LICENSE"
        )
        os.remove(os.path.join(os.getcwd(), "test_dir", "LICENSE"))
        os.rmdir(os.path.join(os.getcwd(), "test_dir"))



"""Connectivity service based on MQTT protocol."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from threading import Lock
from time import sleep
from time import time
from typing import Any
from typing import Callable
from typing import List
from typing import Optional

from paho.mqtt import client as mqtt

from wolk import logger_factory
from wolk.interface.connectivity_service import ConnectivityService
from wolk.model.device import Device
from wolk.model.message import Message

MQTT_KEEP_ALIVE_INTERVAL = 90


class MQTTConnectivityService(ConnectivityService):
    """Handle sending and receiving MQTT messages."""

    def __init__(
        self,
        device: Device,
        topics: List[str],
        last_will_message: Message,
        qos: Optional[int] = 0,
        host: Optional[str] = "api-demo.wolkabout.com",
        port: Optional[int] = 8883,
        ca_cert: Optional[str] = None,
    ) -> None:
        """
        Provide the connection to the WolkAbout IoT Platform.

        :param device: Contains device key, password and actuator references
        :type device: Device
        :param topics: List of topics to subscribe to
        :type topics: List[str]
        :param last_will_message: Message in case of unexpected disconnects
        :type last_will_message: Message
        :param qos: Quality of Service for MQTT connection (0, 1, 2)
        :type qos: int or None
        :param host: Address of the MQTT broker
        :type host: str or None
        :param port: Port to which to send messages
        :type port: int or None
        :param ca_cert: Path to certificate file used to encrypt the connection
        :type ca_cert: str or None
        """
        self.device = device
        self.last_will_message = last_will_message
        self.qos = qos
        self.host = host
        self.port = port
        self.connected = False
        self.connected_rc: Optional[int] = None
        self.topics = topics
        self.ca_cert = ca_cert
        self.logger = logger_factory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug(
            f"Device key: {self.device.key} ; "
            f"Device password: {self.device.password} ;"
            f"Actuator references: {self.device.actuator_references} ; "
            f"Last will message: {self.last_will_message} ; "
            f"QoS: {self.qos} ; "
            f"Host: {self.host} ; "
            f"Port: {self.port} ; "
            f"CA certificate: {self.ca_cert}"
        )
        self.client = mqtt.Client(
            client_id=self.device.key, clean_session=True
        )
        self.inbound_message_listener: Callable[
            [Message], None
        ] = lambda message: print("\n\nNo inbound message listener set!\n\n")
        self.timeout: Optional[int] = None
        self.timeout_interval = 10
        self.mutex = Lock()

    def is_connected(self) -> bool:
        """
        Return current connection state.

        :returns: connected
        :rtype: bool
        """
        return self.connected

    def set_inbound_message_listener(
        self, listener: Callable[[Message], None]
    ) -> None:
        """
        Set the callback function to handle inbound messages.

        :param listener: Function that handles inbound messages
        :type listener: Callable[[Message], None]
        """
        self.logger.debug(f"Message listener set to: {listener}")
        self.inbound_message_listener = listener

    def on_mqtt_message(
        self, _client: mqtt.Client, _userdata: Any, message: mqtt.MQTTMessage
    ) -> None:
        """
        Serialize inbound messages and pass them to inbound message listener.

        :param _client: Client that received the message
        :type _client: paho.mqtt.Client
        :param _userdata: Private user data set in Client()
        :type _userdata: str
        :param message: Class with members: topic, payload, qos, retain.
        :type message: paho.mqtt.MQTTMessage
        """
        if not message:
            return
        received_message = Message(message.topic, message.payload)
        if "binary" in received_message.topic:  # To skip printing file binary
            self.logger.debug(
                "Received MQTT message: "  # type: ignore
                f"{received_message.topic} , "
                f"size: {len(received_message.payload)} bytes"
            )
        else:
            self.logger.debug(f"Received MQTT message: {received_message}")
        self.inbound_message_listener(received_message)

    def on_mqtt_connect(
        self,
        _client: mqtt.Client,
        _userdata: Any,
        _flags: int,
        return_code: int,
    ) -> None:
        """
        Handle when the client receives a CONNACK response from the server.

        :param _client: Client that received the message
        :type _client: paho.mqtt.Client
        :param _userdata: private user data set in Client()
        :type _userdata: str
        :param _flags: Response flags sent by the broker
        :type _flags: int
        :param return_code: Connection result
        :type return_code: int
        """
        self.logger.debug(f"Return code: {return_code}")
        if return_code == 0:  # Connection successful

            self.connected = True
            self.connected_rc = 0
            # Subscribing in on_mqtt_connect() means if we lose the connection
            # and reconnect then subscriptions will be renewed.
            if self.topics:
                for topic in self.topics:
                    self.client.subscribe(topic, 2)
            self.logger.debug(f"Connected : {self.connected}")
        elif (
            return_code == 1
        ):  # Connection refused - incorrect protocol version
            self.connected_rc = 1
        elif (
            return_code == 2
        ):  # Connection refused - invalid client identifier
            self.connected_rc = 2
        elif return_code == 3:  # Connection refused - server unavailable
            self.connected_rc = 3
        elif return_code == 4:  # Connection refused - bad username or password
            self.connected_rc = 4
        elif return_code == 5:  # Connection refused - not authorised
            self.connected_rc = 5

    def on_mqtt_disconnect(
        self, _client: mqtt.Client, _userdata: Any, return_code: int
    ) -> None:
        """
        Handle when the client disconnects from the broker.

        :param _client: Client that received the message
        :type _client: paho.mqtt.Client
        :param _userdata: private user data set in Client()
        :type _userdata: str
        :param return_code: Disconnection result
        :type return_code: int

        :raises RuntimeError: Unexpected disconnection
        """
        self.connected = False
        self.connected_rc = return_code
        self.logger.debug(f"Connected : {self.connected}")
        self.logger.debug(f"Return code : {return_code}")
        if return_code not in [0, 5]:
            self.logger.error("Unexpected disconnect!")
            self.logger.info("Attempting to reconnect..")
            self.connect()

    def connect(self) -> bool:
        """
        Establish the connection to the WolkAbout IoT platform.

        Subscribes to all topics defined by device communication protocol.
        Starts a loop to handle inbound messages.

        :returns: Connection state, True if connected, False otherwise
        :rtype: bool
        """
        if self.connected:
            return True

        self.mutex.acquire()

        self.client.on_connect = self.on_mqtt_connect
        self.client.on_disconnect = self.on_mqtt_disconnect
        self.client.on_message = self.on_mqtt_message
        if self.ca_cert:
            try:
                self.client.tls_set(self.ca_cert)
                self.client.tls_insecure_set(True)
            except Exception as exception:
                self.logger.exception(
                    f"Something went wrong when setting TLS: {exception}"
                )
                self.mutex.release()
                return False
        self.client.username_pw_set(self.device.key, self.device.password)
        self.client.will_set(
            self.last_will_message.topic,
            self.last_will_message.payload,
            2,
            False,
        )

        self.logger.debug(
            f"Connecting with parameters: host='{self.host}', "
            f"port={self.port}, ca_cert='{self.ca_cert}', "
            f"username='{self.device.key}', "
            f"password='{self.device.password}'"
        )
        try:
            self.client.connect(
                self.host, self.port, keepalive=MQTT_KEEP_ALIVE_INTERVAL
            )
        except Exception as exception:
            self.logger.exception(
                f"Something went wrong while connecting: {exception}"
            )
            self.mutex.release()
            return False

        self.client.loop_start()

        self.timeout = round(time()) + self.timeout_interval

        while True:

            if round(time()) > self.timeout:
                self.logger.warning("Connection timed out!")
                self.timeout = None
                self.mutex.release()
                return False

            if self.connected_rc is None:
                sleep(0.1)
                continue

            if self.connected_rc == 0:
                self.logger.info("Connected!")
                self.timeout = None
                self.connected_rc = None
                break

            if self.connected_rc == 1:
                self.logger.warning(
                    "Connection refused - incorrect protocol version"
                )
                self.connected_rc = None
                self.timeout = None
                self.mutex.release()
                return False

            if self.connected_rc == 2:
                self.logger.warning(
                    "Connection refused - invalid client identifier"
                )
                self.connected_rc = None
                self.timeout = None
                self.mutex.release()
                return False

            if self.connected_rc == 3:
                self.logger.warning("Connection refused - server unavailable")
                self.timeout = None
                self.connected_rc = None
                self.mutex.release()
                return False

            if self.connected_rc == 4:
                self.logger.warning(
                    "Connection refused - bad username or password"
                )
                self.connected_rc = None
                self.timeout = None
                self.mutex.release()
                return False

            if self.connected_rc == 5:
                self.logger.warning("Connection refused - not authorised")
                self.connected_rc = None
                self.timeout = None
                self.mutex.release()
                return False

            if self.connected_rc not in list(range(6)):
                self.logger.warning("Unknown retun code")
                self.connected_rc = None
                self.timeout = None
                self.mutex.release()
                return False

        self.logger.debug(f"Subscribing to topics: {self.topics}")
        for topic in self.topics:
            self.client.subscribe(topic, 2)
        self.mutex.release()
        return True

    def disconnect(self) -> None:
        """Disconnects the device from the WolkAbout IoT Platform."""
        self.logger.debug("Disconnecting")
        if self.is_connected():
            self.publish(self.last_will_message)
        self.client.loop_stop()
        self.client.disconnect()

    def publish(self, message: Message) -> bool:
        """
        Publish serialized data to WolkAbout IoT Platform.

        :param message: Message to be published
        :type message: Message

        :returns: result
        :rtype: bool
        """
        if not self.connected:
            self.logger.warning(
                f"Not connected, unable to publish message: {message}"
            )
            return False
        self.mutex.acquire()

        info = self.client.publish(message.topic, message.payload, self.qos)

        if info.rc == mqtt.MQTT_ERR_SUCCESS:
            self.logger.debug(f"Published message: {message}")
            self.mutex.release()
            return True

        if info.is_published():
            self.logger.debug(f"Published message: {message}")
            self.mutex.release()
            return True

        self.logger.warning(f"Failed to publish message: {message}")
        self.mutex.release()
        return False



"""Stub function for setting new configuration values."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Dict
from typing import Union


def handle_configuration(
    configuration: Dict[str, Union[int, float, bool, str]]
) -> None:
    """
    Change device's configuration options.

    Must be implemented as non blocking.
    Must be implemented as thread safe.

    :param configuration: Configuration options as reference:value pairs
    :type configuration: dict
    """
    raise NotImplementedError()



"""Actuator command received from WolkAbout IoT Platform."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from typing import Union


@dataclass
class ActuatorCommand:
    """
    Actuator command for reference with value.

    :ivar reference: What actuator is the command for
    :vartype reference: str
    :ivar value: Value to be set
    :vartype value: Union[bool, int, float, str]
    """

    reference: str
    value: Union[bool, int, float, str]



"""
Module that provides connection to WolkAbout IoT Platform.

To start publishing data to the platform
create an instance of Device class with credentials obtained from the platform
and pass it to an instance of WolkConnect class.

For more information about module features visit:
https://github.com/Wolkabout/WolkConnect-Python/tree/master/examples/full_feature_set
"""
__version__ = "4.1.1"
from .interface.actuation_handler import handle_actuation
from .interface.actuator_status_provider import get_actuator_status
from .interface.configuration_handler import handle_configuration
from .interface.configuration_provider import get_configuration
from .interface.connectivity_service import ConnectivityService
from .interface.firmware_handler import FirmwareHandler
from .interface.message_deserializer import MessageDeserializer
from .interface.message_factory import MessageFactory
from .interface.message_queue import MessageQueue
from .logger_factory import logging_config
from .model.actuator_command import ActuatorCommand
from .model.state import State
from .model.actuator_status import ActuatorStatus
from .model.alarm import Alarm
from .model.configuration_command import ConfigurationCommand
from .model.device import Device
from .model.file_management_error_type import FileManagementErrorType
from .model.file_management_status import FileManagementStatus
from .model.file_management_status_type import FileManagementStatusType
from .model.file_transfer_package import FileTransferPackage
from .model.firmware_update_error_type import FirmwareUpdateErrorType
from .model.firmware_update_status import FirmwareUpdateStatus
from .model.firmware_update_status_type import FirmwareUpdateStatusType
from .model.message import Message
from .repeating_timer import RepeatingTimer
from .model.sensor_reading import SensorReading
from .os_file_management import OSFileManagement
from .os_firmware_update import OSFirmwareUpdate
from .wolk_connect import WolkConnect
from .wolkabout_protocol_message_factory import WolkAboutProtocolMessageFactory
from .wolkabout_protocol_message_deserializer import (
    WolkAboutProtocolMessageDeserializer,
)
from .mqtt_connectivity_service import MQTTConnectivityService
from .message_deque import MessageDeque


__all__ = [
    "ActuatorCommand",
    "State",
    "ActuatorStatus",
    "Alarm",
    "ConfigurationCommand",
    "Device",
    "FileTransferPackage",
    "FirmwareUpdateStatus",
    "FirmwareUpdateStatusType",
    "FirmwareUpdateErrorType",
    "FileManagementStatus",
    "FileManagementStatusType",
    "FileManagementErrorType",
    "Message",
    "SensorReading",
    "handle_actuation",
    "get_actuator_status",
    "handle_configuration",
    "get_configuration",
    "ConnectivityService",
    "FirmwareHandler",
    "MessageDeserializer",
    "MessageFactory",
    "MessageQueue",
    "OSFileManagement",
    "OSFirmwareUpdate",
    "logging_config",
    "RepeatingTimer",
    "WolkConnect",
    "WolkAboutProtocolMessageFactory",
    "WolkAboutProtocolMessageDeserializer",
    "MQTTConnectivityService",
    "MessageDeque",
    "__version__",
]



"""Enables firmware update for device."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Callable

from wolk.interface.firmware_handler import FirmwareHandler
from wolk.model.firmware_update_status import FirmwareUpdateStatus


class FirmwareUpdate(ABC):
    """
    Firmware Update enabler.

    Responsible for supervising firmware installation and reporting current
    firmware installation status and version.
    """

    @abstractmethod
    def __init__(
        self,
        firmware_handler: FirmwareHandler,
        status_callback: Callable[[FirmwareUpdateStatus], None],
    ) -> None:
        """
        Enable firmware update for device.

        :param firmware_handler: Install firmware and provide current version
        :type firmware_handler: FirmwareHandler
        :param status_callback: Reporting firmware update status
        :type status_callback: Callable[[FirmwareUpdateStatus], None]
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_install(self, file_path: str) -> None:
        """
        Handle received firmware installation command.

        :param file_path: Firmware file to install
        :type file_path: str
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_abort(self) -> None:
        """Handle received firmware installation abort command."""
        raise NotImplementedError()

    @abstractmethod
    def report_result(self) -> None:
        """Report the results of the firmware update process."""
        raise NotImplementedError()

    @abstractmethod
    def get_current_version(self) -> str:
        """
        Return device's current firmware version.

        :returns: Firmware version
        :rtype: str
        """
        raise NotImplementedError()



"""Information about the current status of the firmware update process."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from dataclasses import field
from typing import Optional

from wolk.model.file_management_error_type import FileManagementErrorType
from wolk.model.file_management_status_type import FileManagementStatusType


@dataclass
class FileManagementStatus:
    """
    Contains the status of the file management process.

    :ivar status: The status of the file management process
    :ivartype status: FileManagementStatusType
    :ivar error: The type of error that occurred
    :ivartype error: FileManagementErrorType or None
    """

    status: FileManagementStatusType
    error: Optional[FileManagementErrorType] = field(default=None)



"""Repeating timer that calls a fucntion every set interval."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from threading import Timer


class RepeatingTimer:
    """
    Repeating timer calls function with arguments every interval.

    :ivar args: list of arguments
    :vartype args: list
    :ivar f: function to call
    :vartype f: function
    :ivar interval: number of seconds after which to call function
    :vartype interval: int
    :ivar kwargs: keyword arguments
    :vartype kwargs: kwargs
    :ivar timer: timer
    :vartype timer: threading.Timer
    """

    def __init__(self, interval, f, *args, **kwargs):  # type: ignore
        r"""
        Create a repeating timer.

        :param interval: number of seconds after which to call function
        :type interval: int
        :param f: function to call
        :type f: function
        :param \*args: list of arguments
        :type \*args: list
        :param \*\*kwargs: keyword arguments
        :type \*\*kwargs: kwargs
        """
        self.interval = interval
        self.f = f
        self.args = args
        self.kwargs = kwargs

        self.timer = None

    def callback(self) -> None:
        """Call function and start timer."""
        self.f(*self.args, **self.kwargs)
        self.start()

    def cancel(self) -> None:
        """Cancel timer."""
        self.timer.cancel()

    def start(self) -> None:
        """Start timer."""
        self.timer = Timer(self.interval, self.callback)
        self.timer.start()



"""Stub function for providing current device actuator status."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Optional
from typing import Tuple
from typing import Union

from wolk.model.state import State


def get_actuator_status(
    reference: str,
) -> Tuple[State, Optional[Union[bool, int, float, str]]]:
    """
    Get current actuator status identified by reference.

    Reads the status of actuator from the device
    and returns it as a tuple containing the actuator state and current value.

    If unable to get current actuator value,
    return `(State.ERROR, None)`.

    Must be implemented as non blocking.
    Must be implemented as thread safe.

    :param reference: Actuator reference
    :type reference: str
    :returns: (state, value)
    :rtype: (State, bool or int or float or str or None)
    """
    raise NotImplementedError()
